using System;
using System.IO;
using System.Collections;


namespace PTP.Core
{
  /// <summary>
  /// PIMA 15740:2000: DeviceInfo Dataset. This dataset is used to hold the description
  /// information for a device. The Initiator can obtain this dataset from the Responder
  /// without opening a session with the device. This dataset holds data that describes
  /// the device and its capabilities.
  /// </summary>
  public class DeviceInfo
  {
    /// <summary>
    /// PIMA 15740:2000: Highest version of the standard that the device can support.
    /// This represents the standard version expressed in hundredths
    /// (e.g. 1.32 would be stored as 132).
    /// </summary>
    public uint StandardVersion;

    /// <summary>
    /// PIMA 15740:2000: Provides the context for interpretation of any vendor extensions used
    /// by this device. If no extensions are supported, this field shall be set to 0x00000000.
    /// If vendor-specific codes of any type are used, this field is mandatory,
    /// and should not be set to 0x00000000.
    /// </summary>
    public uint VendorExtensionID;

    /// <summary>
    /// PIMA 15740:2000: The vendor-specific version number of extensions that are
    /// supported. This shall be expressed in hundredths (e.g. 1.32 would be stored as 132).
    /// </summary>
    public uint VendorExtensionVersion;

    /// <summary>
    /// PIMA 15740:2000: An optional string used to hold a human-readable
    /// description of the VendorExtensionID. This field should only be used for
    /// informational purposes, and not as the context for the interpretation of vendor-extensions.
    /// </summary>
    public string VendorExtensionDesc;
    
    /// <summary>
    /// PIMA 15740:2000: An optional field used to hold the functional mode. This field
    /// controls whether the device is in an alternate mode that provides
    /// a different set of capabilities (i.e. supported operations, events, etc.)
    /// If the device only supports one mode, this value should always be zero.
    /// 0x0000 = Standard Mode, 0x0001 = Sleep State, ...
    /// </summary>
    public uint FunctionalMode;

    /// <summary>
    /// PIMA 15740:2000: This field is an array of OperationCodes representing operations that
    /// the device is currently supporting, given the FunctionalMode indicated.
    /// </summary>
    public uint[] OperationsSupported;

    /// <summary>
    /// PIMA 15740:2000: This field is an array of EventCodes representing the events that
    /// are currently generated by the device in appropriate situations,
    /// given the FunctionalMode indicated.
    /// </summary>
    public uint[] EventsSupported;

    /// <summary>
    /// PIMA 15740:2000: This field is an array of DevicePropCodes representing
    /// DeviceProperties that are currently exposed for reading and/or modification,
    /// given the FunctionalMode indicated.
    /// </summary>
    public uint[] DevicePropertiesSupported;

    /// <summary>
    /// PIMA 15740:2000: The list of data formats in ObjectFormatCode form that
    /// the device can create using an InitiateCapture operation and/or
    /// an InitiateOpenCapture operation, given the FunctionalMode indicated...
    /// </summary>
    public uint[] CaptureFormats;

    /// <summary>
    /// PIMA 15740:2000: The list of image formats in ObjectFormatCode form that
    /// the device supports in order of highest preference to lowest preference.
    /// Support for an image format refers to the ability to interpret image file
    /// contents according to that formats specifications, for display and/or
    /// manipulation purposes. For image output devices, this field represents
    /// the image formats that the output device is capable of outputting...
    /// </summary>
    public uint[] ImageFormats;

    /// <summary>
    /// PIMA 15740:2000: An optional human-readable string used to hold the
    /// Responders manufacturer.
    /// </summary>
    public string Manufacturer;

    /// <summary>
    /// PIMA 15740:2000: An optional human-readable string used to communicate
    /// the Responders model name.
    /// </summary>
    public string Model;

    /// <summary>
    /// PIMA 15740:2000: An optional string used to communicate the Responders
    /// firmware or software version in a vendor-specific way
    /// </summary>
    public string DeviceVersion;

    /// <summary>
    /// PIMA 15740:2000: An optional string used to communicate the Responders serial number,
    /// which is defined as a unique value among all devices sharing identical Model and
    /// Device Version fields. If unique serial numbers are not supported, this field
    /// shall be set to the empty string. The presence of a non-null string in
    /// the SerialNumber field for one device infers that this field is non-zero and
    /// unique among all devices of that model and version.
    /// </summary>
    public string SerialNumber;

    public DeviceInfo(Stream bytes)
    {
      //FIELD_NAME FIELD_ORDER SIZE DATATYPE 
      //StandardVersion 1 2 UINT16
      this.StandardVersion = DataUtil.ReadUINT16(bytes);
      //VendorExtensionID 2 4 UINT32
      this.VendorExtensionID = DataUtil.ReadUINT32(bytes);
      //VendorExtensionVersion 3 2 UINT16
      this.VendorExtensionVersion = DataUtil.ReadUINT16(bytes);
      //VendorExtensionDesc 4 Variable String
      this.VendorExtensionDesc = DataUtil.ReadString(bytes);
      //FunctionalMode 5 2 UINT16
      this.FunctionalMode = DataUtil.ReadUINT16(bytes);
      //OperationsSupported 6 Variable OperationCode Array
      this.OperationsSupported = DataUtil.ReadUINT16ArrayToMemory(bytes);
      //EventsSupported 7 Variable EventCode Array
      this.EventsSupported = DataUtil.ReadUINT16ArrayToMemory(bytes);
      //DevicePropertiesSupported 8 Variable DevicePropCode Array
      this.DevicePropertiesSupported = DataUtil.ReadUINT16ArrayToMemory(bytes);
      //CaptureFormats 9 Variable ObjectFormatCode Array
      this.CaptureFormats = DataUtil.ReadUINT16ArrayToMemory(bytes);
      //ImageFormats 10 Variable ObjectFormatCode Array
      this.ImageFormats = DataUtil.ReadUINT16ArrayToMemory(bytes);
      //Manufacturer 11 Variable String
      this.Manufacturer = DataUtil.ReadString(bytes);
      //Model 12 Variable String
      this.Model = DataUtil.ReadString(bytes);
      //DeviceVersion 13 Variable String
      this.DeviceVersion = DataUtil.ReadString(bytes);
      //SerialNumber 14 Variable String
      this.SerialNumber = DataUtil.ReadString(bytes);
    }
  }

  public class StorageInfo
  {
    /// <summary>
    /// PIMA 15740:2000: The code that identifies the type of storage,
    /// particularly whether the store is inherently random-access or read-only memory,
    /// and whether it is fixed or removable media.
    /// </summary>
    public uint StorageType;

    /// <summary>
    /// PIMA 15740:2000: This optional code indicates the type of filesystem present
    /// on the device. This field may be used to determine the filenaming convention
    /// used by the storage device, as well as to determine whether support
    /// for a hierarchical system is present. If the storage device is DCF-conformant,
    /// it shall indicate so here.
    /// </summary>
    public uint FilesystemType;

    /// <summary>
    /// PIMA 15740:2000: This field indicates whether the store is read-write or read-only.
    /// If the store is read-only, deletion may or may not be allowed. The allowed values are
    /// described in the following table. Read-Write is only valid if the StorageType is
    /// non-ROM, as described in the StorageType field above.
    /// </summary>
    public uint AccessCapability;

    /// <summary>
    /// PIMA 15740:2000: This is an optional field that indicates the total storage capacity of
    /// the store in bytes. If this field is unused, it should report 0xFFFFFFFF.
    /// </summary>
    public ulong MaxCapacity;

    /// <summary>
    /// PIMA 15740:2000: The amount of free space that is available in the store in bytes.
    /// If this value is not useful for the device, it may set this field to 0xFFFFFFFF and
    /// rely upon the FreeSpaceInImages field instead.
    /// </summary>
    public ulong FreeSpaceInBytes;

    /// <summary>
    /// PIMA 15740:2000: The number of images that may still be captured into this store
    /// according to the current image capture settings of the device. If the device does not
    /// implement this capability, this field should be set to 0xFFFFFFFF. This field may be
    /// used for devices that do not report FreeSpaceInBytes, or the two fields may be
    /// used in combination.
    /// </summary>
    public uint FreeSpaceInImages;

    /// <summary>
    /// PIMA 15740:2000: An optional field that may be used for a human-readable text
    /// description of the storage device. This should be used for storage-type specific
    /// information as opposed to volume-specific information. Examples would be
    /// "Type I Compact Flash," or "3.5-inch 1.44 MB Floppy". If unused, this field
    /// should be set to the empty string.
    /// </summary>
    public string StorageDescription;

    /// <summary>
    /// PIMA 15740:2000: An optional field that may be used to hold the volume label of
    /// the storage device, if such a label exists and is known. If unused, this field
    /// should be set to the empty string.
    /// </summary>
    public string VolumeLabel;

    public StorageInfo(Stream s)
    {
      //FIELD_NAME FIELD_ORDER SIZE DATATYPE 
      // StorageType 1 2 UINT16
      StorageType = DataUtil.ReadUINT16(s);
      // FilesystemType 2 2 UINT16
      FilesystemType = DataUtil.ReadUINT16(s);
      // AccessCapability 3 2 UINT16
      AccessCapability = DataUtil.ReadUINT16(s);
      // MaxCapacity 4 8 UINT64
      AccessCapability = DataUtil.ReadUINT64(s);
      // FreeSpaceInBytes 5 8 UINT64
      FreeSpaceInBytes = DataUtil.ReadUINT64(s);
      // FreeSpaceInImages 6 4 UINT32
      FreeSpaceInImages = DataUtil.ReadUINT32(s);
      // StorageDescription 7 Variable String
      StorageDescription = DataUtil.ReadString(s);
      // VolumeLabel 8 Variable String
      VolumeLabel = DataUtil.ReadString(s);
    }
  }
}
